# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
from cil.optimisation.algorithms import Algorithm
import warnings

class GD(Algorithm):
    ''' 
    
        Gradient Descent algorithm
        
        
    '''

    def __init__(self, initial=None, objective_function=None, step_size =None, **kwargs):
        '''GD algorithm creator
        
        initialisation can be done at creation time if all 
        proper variables are passed or later with set_up
        
        :param initial: initial guess
        :param objective_function: objective function to be minimised
        :param step_size: step size for gradient descent iteration
        :param alpha: optional parameter to start the backtracking algorithm, default 1e6
        :param beta: optional parameter defining the reduction of step, default 0.5.
                    It's value can be in (0,1)
        :param rtol: optional parameter defining the relative tolerance comparing the 
                     current objective function to 0, default 1e-5, see numpy.isclose
        :param atol: optional parameter defining the absolute tolerance comparing the 
                     current objective function to 0, default 1e-8, see numpy.isclose
        '''
        super(GD, self).__init__(**kwargs)
        if kwargs.get('x_init', None) is not None:
            if initial is None:
                warnings.warn('The use of the x_init parameter is deprecated and will be removed in following version. Use initial instead',
                   DeprecationWarning, stacklevel=4)
                initial = kwargs.get('x_init', None)
            else:
                raise ValueError('{} received both initial and the deprecated x_init parameter. It is not clear which one we should use.'\
                    .format(self.__class__.__name__))

        self.alpha = kwargs.get('alpha' , 1e6)
        self.beta = kwargs.get('beta', 0.5)
        self.rtol = kwargs.get('rtol', 1e-5)
        self.atol = kwargs.get('atol', 1e-8)
        if initial is not None and objective_function is not None :
            self.set_up(initial=initial, objective_function=objective_function, step_size=step_size)
    
    def set_up(self, initial, objective_function, step_size):
        '''initialisation of the algorithm
        
        :param initial: initial guess
        :param objective_function: objective function to be minimised
        :param step_size: step size'''
        print("{} setting up".format(self.__class__.__name__, ))
            
        self.x = initial.copy()
        self.objective_function = objective_function
        

        if step_size is None:
            self.k = 0
            self.update_step_size = True
            self.x_armijo = initial.copy()
            # self.rate = self.armijo_rule() * 2
            # print (self.rate)
        else:
            self.step_size = step_size
            self.update_step_size = False
        
        
        self.update_objective()

        self.x_update = initial.copy()

        self.configured = True
        print("{} configured".format(self.__class__.__name__, ))

    def update(self):
        '''Single iteration'''
        
        self.objective_function.gradient(self.x, out=self.x_update)
        
        if self.update_step_size:
            # the next update and solution are calculated within the armijo_rule
            self.step_size = self.armijo_rule()
        else:
            self.x_update *= -self.step_size
            self.x += self.x_update
        
    

    def update_objective(self):
        self.loss.append(self.objective_function(self.x))

    def armijo_rule(self):
        '''Applies the Armijo rule to calculate the step size (step_size)

        https://projecteuclid.org/download/pdf_1/euclid.pjm/1102995080

        The Armijo rule runs a while loop to find the appropriate step_size by starting
        from a very large number (alpha). The step_size is found by dividing alpha by 2 
        in an iterative way until a certain criterion is met. To avoid infinite loops, we 
        add a maximum number of times the while loop is run. 

        This rule would allow to reach a minimum step_size of 10^-alpha. 
        
        if
        alpha = numpy.power(10,gamma)
        delta = 3
        step_size = numpy.power(10, -delta)
        with armijo rule we can get to step_size from initial alpha by repeating the while loop k times 
        where 
        alpha / 2^k = step_size
        10^gamma / 2^k = 10^-delta 
        2^k = 10^(gamma+delta)
        k = gamma+delta / log10(2) \\approx 3.3 * (gamma+delta)

        if we would take by default delta = gamma
        kmax = numpy.ceil ( 2 * gamma / numpy.log10(2) )
        kmax = numpy.ceil (2 * numpy.log10(alpha) / numpy.log10(2))

        '''
        f_x = self.objective_function(self.x)
        if not hasattr(self, 'x_update'):
            self.x_update = self.objective_function.gradient(self.x)
        
        while self.k < self.kmax:
            # self.x - alpha * self.x_update
            self.x_update.multiply(self.alpha, out=self.x_armijo)
            self.x.subtract(self.x_armijo, out=self.x_armijo)
            
            f_x_a = self.objective_function(self.x_armijo)
            sqnorm = self.x_update.squared_norm()
            if f_x_a - f_x <= - ( self.alpha/2. ) * sqnorm:
                self.x.fill(self.x_armijo)
                break
            else:
                self.k += 1.
                # we don't want to update kmax
                self._alpha *= self.beta

        if self.k == self.kmax:
            raise ValueError('Could not find a proper step_size in {} loops. Consider increasing alpha.'.format(self.kmax))
        return self.alpha
    
    @property
    def alpha(self):
        return self._alpha
    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha
        self.kmax = numpy.ceil (2 * numpy.log10(alpha) / numpy.log10(2))
        self.k = 0
        
    def should_stop(self):
        return self.max_iteration_stop_cryterion() or \
            numpy.isclose(self.get_last_objective(), 0., rtol=self.rtol, 
                atol=self.atol, equal_nan=False)





# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cil.optimisation.functions import L1Norm, ScaledFunction, \
                                        LeastSquares, L2NormSquared, \
                                        KullbackLeibler, ZeroFunction, ConstantFunction
from cil.optimisation.operators import IdentityOperator                                        
from cil.framework import ImageGeometry

import unittest
import numpy
from numbers import Number

''' Here we test SumFunction class for different function

L2Norm, L1Norm, KullbackLeibler, ZeroFunction, ConstantFunction, Scalar

for call method
for gradient method



'''



class TestFunction(unittest.TestCase):
    
    def assertBlockDataContainerEqual(self, container1, container2):
        print ("assert Block Data Container Equal")
        self.assertTrue(issubclass(container1.__class__, container2.__class__))
        for col in range(container1.shape[0]):
            if issubclass(container1.get_item(col).__class__, DataContainer):
                print ("Checking col ", col)
                self.assertNumpyArrayEqual(
                    container1.get_item(col).as_array(), 
                    container2.get_item(col).as_array()
                    )
            else:
                self.assertBlockDataContainerEqual(container1.get_item(col),container2.get_item(col))
    
    def assertNumpyArrayEqual(self, first, second):
        res = True
        try:
            numpy.testing.assert_array_equal(first, second)
        except AssertionError as err:
            res = False
            print(err)
        self.assertTrue(res)

    def assertNumpyArrayAlmostEqual(self, first, second, decimal=6):
        res = True
        try:
            numpy.testing.assert_array_almost_equal(first, second, decimal)
        except AssertionError as err:
            res = False
            print(err)
            print("expected " , second)
            print("actual " , first)

        self.assertTrue(res)
        
    def test_SumFunction(self):
        
        M, N, K = 3,4,5
        ig = ImageGeometry(M, N, K)
        
        tmp = ig.allocate('random', seed=1)
        b   = ig.allocate('random', seed=2)
        eta = ig.allocate(0.1)
        
        operator = IdentityOperator(ig)

        scalar = 0.25
        f1 = L2NormSquared()
        f2 = L1Norm()
        f3 = scalar * L2NormSquared()
        f4 = scalar * L1Norm()
        f5 = scalar * L2NormSquared(b=b)
        f6 = scalar * L1Norm(b=b)  
        f7 = ZeroFunction()
        f8 = 5 *  ConstantFunction(10)             
        f9 = LeastSquares(operator, b, c=scalar)
        f10 = 0.5*KullbackLeibler(b=b,eta = eta)
        f11 = KullbackLeibler(b=b, eta =eta)
        f12 = 10
        
#        f10 = 0.5 * MixedL21Norm()
#        f11 = IndicatorBox(lower=0)
        
        list1 = [f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12]
                
        print('###################  Check sum of two functions ################## \n')
        
        for func in list1:
               
            
            # check sum of two functions   
            
            if isinstance(func, ScaledFunction):
                type_fun = ' scalar * ' + type(func.function).__name__
            else:    
                type_fun = type(func).__name__
                
            if isinstance(func, Number):
                tmp_fun_eval = func
            else:
                tmp_fun_eval = func(tmp)                
                             
            sumf = f1 + func           
            self.assertNumpyArrayAlmostEqual(sumf(tmp), f1(tmp) + tmp_fun_eval )
            print('{} = ( {} + {} ) is OK'.format(type(sumf).__name__, type(f1).__name__, type_fun))
            
            sumf1 = func + f1 
            self.assertNumpyArrayAlmostEqual(sumf1(tmp), tmp_fun_eval + f1(tmp))
            print('Checking commutative')
            print('{} + ( {} + {} ) is OK\n'.format(type(sumf1).__name__, type_fun, type(f1).__name__))
            
        print('###################  Check Lispchitz constant ################## \n')
        
        for i,func in enumerate(list1):
            
            if isinstance(func, ScaledFunction):
                type_fun = ' scalar * ' + type(func.function).__name__
            else:    
                type_fun = type(func).__name__            
               
            try:
                # check Lispchitz sum of two functions  
                print ("i", i,func.__class__.__name__)
                if isinstance(func, Number):
                    tmp_fun_L = 0
                else:
                    tmp_fun_L = func.L           
                
                sumf = f1 + func   
                
                try:
                    sumf.L==f1.L + tmp_fun_L
                except TypeError:
                    print('Function {} has L = None'.format(type_fun))
            except ValueError as nie:
                print (func.__class__.__name__, nie)
                
        print('\n###################  Check Gradient ################## \n')   
              
              
        for func in list1:
              
            if isinstance(func, ScaledFunction):
                type_fun = ' scalar * ' + type(func.function).__name__
            else:    
                type_fun = type(func).__name__
                                          
            sumf = f1 + func
            # check gradient          
            try:
                if isinstance(func, Number):
                    tmp_fun_gradient = 0
                else:
                    tmp_fun_gradient = func.gradient(tmp)   
                    
                self.assertNumpyArrayAlmostEqual(sumf.gradient(tmp).as_array(), (f1.gradient(tmp) + tmp_fun_gradient).as_array())
            except NotImplementedError:
                print("{} is not differentiable".format(type_fun))
                
        print('\n###################  Check Gradient Out ################## \n')  
         
        out_left = ig.allocate()
        out_right1 = ig.allocate()
        out_right2 = ig.allocate()  
            
        for i, func in enumerate(list1):               
                
            if isinstance(func, ScaledFunction):
                type_fun = ' scalar * ' + type(func.function).__name__
            else:    
                type_fun = type(func).__name__
            
            sumf = f1 + func
            
                                    
            # check gradient out    
            try:
                
                
                if isinstance(func, Number):
                    tmp_fun_gradient_out = 0
                else:
                    func.gradient(tmp, out = out_right2) 
                    tmp_fun_gradient_out = out_right2.as_array()
                    
                #print('Check {} + {}\n'.format(type(f1).__name__, type_fun))
                sumf.gradient(tmp, out = out_left)
                f1.gradient(tmp, out = out_right1)   
                self.assertNumpyArrayAlmostEqual(out_left.as_array(), out_right1.as_array() + tmp_fun_gradient_out)
            except NotImplementedError:
                print("{} is not differentiable".format(type_fun))  
    def test_ConstantFunction(self):

        k = ConstantFunction(constant=1)
        ig = ImageGeometry(1,2,3)
        x = ig.allocate(2)

        grad = k.gradient(x)
        out = ig.allocate(-1)

        k.gradient(x, out=out)
        #out.fill(-3)
        
        self.assertNumpyArrayEqual(numpy.zeros(x.shape), grad.as_array())
        
        self.assertNumpyArrayEqual(out.as_array(), grad.as_array())
                
    def test_SumFunctionScalar(self):      
        numpy.random.seed(1)
        M, N, K = 3,4,5
        ig = ImageGeometry(M, N, K)
        
        tmp = ig.allocate('random')
        b = ig.allocate('random')
        
        scalar = 0.25
        f1 =  scalar * L2NormSquared(b=b)
        f2 = 5
           
        f = f1 + f2
        print(f)
        
        g = f2 + f1
        print(g)

        tau = 0.03
        
        # check call method       
        res1 = f(tmp)
        res2 = f1(tmp) + f2
        self.assertAlmostEqual(res1, res2)
        
        # check gradient
        res1 = f.gradient(tmp)
        res2 = f1.gradient(tmp)
        self.assertNumpyArrayAlmostEqual(res1.as_array(), res2.as_array())
        
        # check gradient with out
        out1 = tmp*0
        out2 = tmp*0
        f.gradient(tmp, out=out1)
        f1.gradient(tmp, out=out2)
        self.assertNumpyArrayAlmostEqual(out1.as_array(), out2.as_array())
        

        res1 = f.proximal(tmp, tau)
        res2 = f1.proximal(tmp, tau)
        
        # proximal of sum of function with scalar = proximal of function
        self.assertNumpyArrayAlmostEqual(res1.as_array(), res2.as_array())
        
        # proximal  with out of sum of function with scalar = proximal of function
        res1_out = ig.allocate()
        res2_out = ig.allocate()         
        f.proximal(tmp, tau, out = res1_out)
        f1.proximal(tmp, tau, out = res2_out)             
        self.assertNumpyArrayAlmostEqual(res1_out.as_array(), res2_out.as_array())

        res1 = f.proximal_conjugate(tmp, tau)
        res2 = f1.proximal_conjugate(tmp, tau)
        
        # proximal of sum of function with scalar = proximal of function
        self.assertNumpyArrayAlmostEqual(res1.as_array(), res2.as_array())
        
        # proximal  with out of sum of function with scalar = proximal of function
        res1_out = ig.allocate()
        res2_out = ig.allocate()         
        f.proximal_conjugate(tmp, tau, out = res1_out)
        f1.proximal_conjugate(tmp, tau, out = res2_out)             
        self.assertNumpyArrayAlmostEqual(res1_out.as_array(), res2_out.as_array())        
            
        
def test_ConstantFunction(self):      
        
        M, N, K = 3,4,5
        ig = ImageGeometry(M, N, K)
        
        tmp = ig.allocate('random_int')
        
        constant = 10
        f = ConstantFunction(constant)
        
        # check call
        res1 = f(tmp)
        self.assertAlmostEqual(res1, constant)
        
        # check gradient with and without out
        res1 = f.gradient(tmp)
        out = ig.allocate()
        self.assertNumpyArrayAlmostEqual(res1.as_array(), out)
        
        out1 = ig.allocate()
        f.gradient(tmp, out=out1)
        self.assertNumpyArrayAlmostEqual(res1.as_array(), out1)
        
        # check convex conjugate        
        res1 = f.convex_conjugate(tmp)
        res2 = tmp.maximum(0).sum()
        self.assertNumpyArrayAlmostEqual(res1.as_array(), res2.as_array())
        
        # check proximal 
        tau = 0.4
        res1 = f.proximal(tmp, tau)
        self.assertNumpyArrayAlmostEqual(res1.as_array(), tmp.as_array()) 
            
        
        
        # check call
        
        
        
     
        
                
if __name__ == '__main__':
#    
    t = TestFunction()
    t.test_SumFunction()
#    t.test_SumFunctionScalar()

                



# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy
from cil.utilities import dataexample
from cil.utilities import noise
import os, sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from testclass import CCPiTestClass
import unittest

class TestTestData(CCPiTestClass):
    def test_noise_gaussian(self):
        camera = dataexample.CAMERA.get()
        noisy_camera = noise.gaussian(camera, seed=1)
        norm = (camera - noisy_camera).norm()
        self.assertAlmostEqual(norm, 48.881268, places=4)

    def test_load_CAMERA(self):

        
        res = False
        try:
            image = dataexample.CAMERA.get()
            if (image.shape[0] == 512) and (image.shape[1] == 512):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)


    def test_load_BOAT(self):
        
        res = False
        try:
            image = dataexample.BOAT.get()
            if (image.shape[0] == 512) and (image.shape[1] == 512):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)

    def test_load_PEPPERS(self):
        
        res = False
        try:
            image = dataexample.PEPPERS.get()
            if (image.shape[0] == 3) and (image.shape[1] == 512) and (image.shape[2] == 512):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)

    def test_load_RAINBOW(self):
        
        res = False
        try:
            image = dataexample.RAINBOW.get()
            if (image.shape[0] == 3) and (image.shape[1] == 1353) and (image.shape[2] == 1194):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)
    def test_load_RESOLUTION_CHART(self):
        
        res = False
        try:
            image = dataexample.RESOLUTION_CHART.get()
            if (image.shape[0] == 256) and (image.shape[1] == 256):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)

    def test_load_SIMPLE_PHANTOM_2D(self):
        res = False
        try:
            image = dataexample.SIMPLE_PHANTOM_2D.get()
            if (image.shape[0] == 512) and (image.shape[1] == 512):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)

    def test_load_SHAPES(self):
        res = False
        try:
            image = dataexample.SHAPES.get()
            if (image.shape[0] == 200) and (image.shape[1] == 300):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)

    def test_load_SYNCHROTRON_PARALLEL_BEAM_DATA(self):
        res = False
        try:
            image = dataexample.SYNCHROTRON_PARALLEL_BEAM_DATA.get()
            if (image.shape[0] == 91) and (image.shape[1] == 135) and\
                (image.shape[2] == 160):
                res = True
            else:
                print("Image dimension mismatch")
        except FileNotFoundError:
            print("File not found")
        except:
            print("Failed to load file")

        self.assertTrue(res)
        



# -*- coding: utf-8 -*-
#   This work is part of the Core Imaging Library (CIL) developed by CCPi 
#   (Collaborative Computational Project in Tomographic Imaging), with 
#   substantial contributions by UKRI-STFC and University of Manchester.

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from cil.framework import DataProcessor, AcquisitionData, ImageData, DataContainer, AcquisitionGeometry, ImageGeometry
import numpy
import warnings


class Binner(DataProcessor):
    r'''Binner processor rebins (downsample) array and returns new geometry.
        
    :param roi: region-of-interest to bin, specified as a dictionary containing tuple (Start, Stop, Step)
    :type roi: dict
    :return: returns an AcquisitionData or ImageData object with an updated AcquisitionGeometry or ImageGeometry
    :rtype: AcquisitionData or ImageData
    '''
    
    '''
    Start inclusive, Stop exclusive
    
    -1 is a shortcut to include all elements along the specified dimension
    
    if only one number is provided, then it is interpreted as Stop
    
    if two numbers are provided, then they are interpreted as Start and Stop
    
    Start = None is equivalent to Start = 0
    Stop = None is equivalent to Stop = number of elements
    Step = None is equivalent to Step = 1
    
    You can specify negative Start and Stop.
    
    If Stop - Start is not multiple of Step, then 
    the resulted dimension will have (Stop - Start) // Step 
    elements, i.e. (Stop - Start) % Step elements will be ignored
    '''

    def __init__(self,
                 roi = None):

        kwargs = {'roi': roi}

        super(Binner, self).__init__(**kwargs)
    

    def check_input(self, data):
        
        if not ((isinstance(data, ImageData)) or 
                (isinstance(data, AcquisitionData))):
            raise TypeError('Processor supports only following data types:\n' +
                            ' - ImageData\n - AcquisitionData')
        elif (data.geometry == None):
            raise ValueError('Geometry is not defined.')
        elif (self.roi == None):
            raise ValueError('Prease, specify roi')
        else:
            return True 
    

    def process(self, out=None):

        data = self.get_input()
        ndim = data.number_of_dimensions

        geometry_0 = data.geometry
        geometry = geometry_0.copy()

        dimension_labels = list(geometry_0.dimension_labels)
        
        if self.roi != None:
            for key in self.roi.keys():
                if key not in data.dimension_labels:
                    raise ValueError('Wrong label is specified for roi, expected {}.'.format(data.dimension_labels))
        
        roi_object = self._construct_roi_object(self.roi, data.shape, dimension_labels)

        
        for key in self.roi.keys():
            idx = data.dimension_labels.index(key)
            n_elements = (roi_object[idx][1] - roi_object[idx][0]) // roi_object[idx][2]
            
            if (isinstance(data, ImageData)):

                if key == 'channel':
                    geometry.channels = n_elements
                    geometry.channel_spacing *= roi_object[idx][2]
                    if n_elements <= 1:
                        dimension_labels.remove('channel')
                elif key == 'vertical':
                    geometry.voxel_num_z = n_elements
                    geometry.voxel_size_z *= roi_object[idx][2]
                    if n_elements <= 1:
                        dimension_labels.remove('vertical')
                elif key == 'horizontal_x':
                    geometry.voxel_num_x = n_elements
                    geometry.voxel_size_x *= roi_object[idx][2]
                    if n_elements <= 1:
                       dimension_labels.remove('horizontal_x')
                elif key == 'horizontal_y':
                    geometry.voxel_num_y = n_elements
                    geometry.voxel_size_y *= roi_object[idx][2]
                    if n_elements <= 1:
                        dimension_labels.remove('horizontal_y')
            
            # if AcquisitionData
            else:
                if key == 'channel':
                    geometry.set_channels(num_channels=n_elements)
                    if n_elements <= 1:
                        dimension_labels.remove('channel')
                elif key == 'angle':
                    shape = (n_elements, roi_object[idx][2])
                    geometry.config.angles.angle_data = geometry_0.config.angles.angle_data[roi_object[idx][0]:(roi_object[idx][0] + n_elements * roi_object[idx][2])].reshape(shape).mean(1)
                    if n_elements <= 1:
                        dimension_labels.remove('angle')
                elif key == 'vertical':
                    if n_elements > 1:
                        geometry.config.panel.num_pixels[1] = n_elements
                    else:
                        geometry = geometry.subset(vertical = (roi_object[idx][1] + roi_object[idx][0]) // 2)
                    geometry.config.panel.pixel_size[1] *= roi_object[idx][2]
                elif key == 'horizontal':
                    geometry.config.panel.num_pixels[0] = n_elements
                    geometry.config.panel.pixel_size[0] *= roi_object[idx][2]
                    if n_elements <= 1:
                        dimension_labels.remove('horizontal')
        
        geometry.dimension_labels = dimension_labels
        
        shape_object = []
        slice_object = []
        for i in range(ndim):
            n_pix = (roi_object[i][1] - roi_object[i][0]) // roi_object[i][2]
            shape_object.append(n_pix)
            shape_object.append(roi_object[i][2])
            slice_object.append(slice(roi_object[i][0], roi_object[i][0] + n_pix * roi_object[i][2]))
        
        shape_object = tuple(shape_object)
        slice_object = tuple(slice_object)

        data_resized = data.as_array()[slice_object].reshape(shape_object)

        mean_order = [-1, 1, 2, 3]

        for i in range(ndim):
            data_resized = data_resized.mean(mean_order[i])
            
        data_binned = geometry.allocate()
        data_binned.fill(numpy.squeeze(data_resized))
        if out == None:
            return data_binned
        else:
            out = data_binned
        

    def _construct_roi_object(self, roi, n_elements, dimension_labels):

        '''
        parse roi input
        here we parse input and calculate requested roi
        '''
        ndim = len(n_elements)
        roi_object = []
        # loop through dimensions
        for i in range(ndim):
            # given dimension number, get corresponding label
            label = dimension_labels[i]
            # '-1' shortcut = include all elements
            if (label in roi.keys()) and (roi[label] != -1):
                # start and step are optional
                if len(roi[label]) == 1:
                    start = 0
                    step = 1
                    if roi[label][0] != None:
                        if roi[label][0] < 0:
                            stop =  n_elements[i]+roi[label][0]
                        else:
                            stop = roi[label][0]
                    else:
                        stop =  n_elements[i]

                elif len(roi[label]) == 2 or len(roi[label]) == 3:
                    
                    if roi[label][0] != None:
                        if roi[label][0] < 0:
                            start = n_elements[i]+roi[label][0]
                        else:
                            start = roi[label][0]
                    else:
                        start = 0

                    if roi[label][1] != None:
                        if roi[label][1] < 0:
                            stop = n_elements[i]+roi[label][1]
                        else:
                            stop = roi[label][1]
                    else: 
                        stop = n_elements[i]
                    
                    if len(roi[label]) == 2:
                        step = 1

                    if len(roi[label]) == 3:
                        if roi[label][2] != None:
                            if roi[label][2] <= 0:
                                raise ValueError('Binning parameter has to be > 0')
                            else:
                                step = roi[label][2]
                        else:
                            step = 1
                else:
                    raise ValueError('roi is exected to have 1, 2 or 3 elements')
            else:
                step = 1
                start = 0
                stop = n_elements[i]
            roi_object.append((start, stop, step))
        return roi_object